(in-package :probgen-mirror-ninja)

(defun make-grid-model (n)
  (let ((size (* n n)))
    (loop for loc from 1 to size
          as north = (if (< (+ loc n) size)
                       (+ loc n)
                       -1)
          as south = (if (> (- loc n) 0)
                       (- loc n)
                       -1)
          
          as east = (if 
                        (and (= (mod loc n) 0)
                             (= (mod (+ loc 1) n) 1))
                      -1
                      (+ loc 1))
          
          as west = (if (= (mod (- loc 1) n) 0)
                      -1
                      (- loc 1))
          when (> north -1)
            collect `(adjacent-north ,loc ,north) into grid-model
          when (> south -1)
            collect `(adjacent-south ,loc ,south) into grid-model
          when (> east -1)
            collect `(adjacent-east ,loc ,east) into grid-model
          when (> west -1)
            collect `(adjacent-west ,loc ,west) into grid-model
          finally (return grid-model))))

(defun make-cells (n)
  (let ((size (* n n)))
    (loop for i from 1 to size
          collect i)))

(defun problem-instance (n)
  (let* ((size (* n n))
         (ninja-zone (loop for i from 1 to n collect i))
         (mirror-zone (loop for i from (1+ (- size n)) to size collect i))
         (ninja-init (nth (random (length ninja-zone))
                          ninja-zone))
         (mirror-init (+ ninja-init 
                         (* n (- n 1))))
         (ninja-goal (1+ (random (floor (/ size 2))))))
  
    `(define (problem prob)
         (:domain mirror-ninja)
       (:objects
        ninja mirror - ninja
        ,@(make-cells n) - location)
       (:init
        (at ninja ,ninja-init)
        (at mirror ,mirror-init)
        ,@(make-grid-model n)
        )
       (:goal (and (at ninja ,ninja-goal))))))

(defun all-permutations (list)
  (cond ((null list) nil)
        ((null (cdr list)) (list list))
        (t (loop for element in list
             append (mapcar (lambda (l) (cons element l))
                            (all-permutations (remove element
                                                      list)))))))
(defun next-step (list)
  (nth (random (length list)) list))

(defun make-step-orderings (n)
  (let* ((size (* n n))
         (steps '(step1 step2 step3))
         step-predecessors)
    (loop for loc from 1 to size 
          do (let (s1 s2 s3
                      (nsteps (copy-list steps)))
               (setf s1 (next-step nsteps))
               (setq nsteps (remove s1 nsteps))
               (setf s2  (next-step nsteps))
               (setq nsteps (remove s2 nsteps))
               (setf s3  (next-step nsteps))
               
               (setf step-predecessors
                     (append `((finished-step step0 ,loc)
                               (predecessor ,s1 step0 ,loc)
                               (predecessor ,s2 ,s1 ,loc)
                               (predecessor ,s3 ,s2 ,loc))
                             step-predecessors))))
    step-predecessors))

(defun make-steps-unordered (n)
  (let* ((size (* n n)))
    (loop for loc from 1 to size 
          append `((finished-step step0 ,loc)
                             (predecessor step1 step0 ,loc)
                             (predecessor step2 step0 ,loc)
                             (predecessor step3 step0 ,loc))
            into step-predecessors
          finally (return step-predecessors))))

(defun problem-instance-goals (n)
  (let* ((size (* n n))
         (ninja-zone (loop for i from 1 to n collect i))
         (mirror-zone (loop for i from (1+ (- size n)) to size collect i))
         (ninja-init 1) 
         (mirror-init 21)
         
         (ninja-goal 25))
  
    `(define (problem prob)
         (:domain mirror-ninja)
       (:objects
        ninja - ninja 
        mirror - mirror
        step0 step1 step2 step3 - step
        ,@(make-cells n) - location)
       (:init
        (at-ninja ninja ,ninja-init)
        (at-mirror mirror ,mirror-init)
        ,@(make-grid-model n)
        ,@(make-step-orderings n)
        )
       (:goal (and (at-ninja ninja ,ninja-goal))))))

(defun problem-instance-unordered (n)
  (let* ((size (* n n))
         (ninja-zone (loop for i from 1 to n collect i))
         (mirror-zone (loop for i from (1+ (- size n)) to size collect i))
         (ninja-init 1) 
         (mirror-init 21)
         (ninja-goal 25))
  
    `(define (problem prob)
         (:domain mirror-ninja)
       (:objects
        ninja - ninja 
        mirror - mirror
        step0 step1 step2 step3 - step
        ,@(make-cells n) - location)
       (:init
        (at-ninja ninja ,ninja-init)
        (at-mirror mirror ,mirror-init)
        ,@(make-grid-model n)
        ,@(make-steps-unordered n)
        )
       (:goal (and (at-ninja ninja ,ninja-goal))))))

(defun mirror-ninja-problems (n problem-num output-directory
                              &key (goal-type :single))
  (loop for i from 1 to problem-num
        as problem = (ecase goal-type
                       (:single (problem-instance n))
                       (:subgoals
                              (problem-instance-goals n)))
        do (with-open-file (str (format nil "~a/prob~a.pddl" 
                                  output-directory i)
                                :direction :output
                                :if-exists :supersede)
             (pprint problem str))))

(defun mirror-ninja-unordered-problems (n problem-num
                                        output-directory
                                        domain-file)
  (let ((base-problem (problem-instance-unordered n))
        (problem-file (format nil "~a/base-problem.pddl"
                                        output-directory))
        plan plans)

    (with-open-file (str problem-file
                         :direction :output
                         :if-exists :supersede)
      (pprint base-problem str))
    (setf plan (fast-downward::run-planner domain-file
                                            problem-file :fd
                                            :output-directory output-directory 
                                            :fd-config :lama))
    
    (format t "~%Plan: ~a" plan)
    (format t "~%Length: ~a" (length plan))

    (loop for i from 1 to problem-num 
          do (let ((start 0) (end 3) 
                   (plan1 (copy-list plan)))
               (loop while (< end (length plan1))
                     do (setf plan1 
                              (append (subseq plan1 0 start)
                                      (shuffle-plan-one-step (subseq plan1 start
                                                                     end))
                                      (subseq plan1 end)))
                        (setf start (+ start 5)
                              end (+ end 5)))
               (push plan1 plans) 
               (format t "~%Plan1: ~a" plan1)
               (format t "~%Length: ~a" (length plan1))))

    (with-open-file (str (format nil "~a/plans-for-unordered.pddl" output-directory)
                         :direction :output
                         :if-exists :supersede)
      (pprint plans str))))

(defun shuffle-plan-one-step (plan)
  (let ((pivot (first plan))
        (swap (random (length (rest plan)))))
    (setf (nth 0 plan) (nth swap (rest plan)))
    (setf (nth swap (rest plan)) pivot)
    plan))
          
    
        

        
    

